                                       
#**************************************************************************** 
#** 
#**  File          : /lua/AreaShield_SuperClass.lua 
#** 
#**  Author        : Resin_Smoker 
#**
#**  Special Thanks: BulletMagnet, Mithy, DeadMG, Seiya, Exavier Macbeth
#** 
#**  Summary       : This script allots a Shield Projection Unit (SPU) to 
#**                  enhance MLU's (Moble Land Units) and Structures with
#**                  personal shields. All friendly units within range that 
#**                  are not on the exclude list can be shielded.                                     
#** 
#**  Copyright     :  Copyright  28 FEB 2010 
#** 
#**************************************************************************** 

#### Localy imported files ####
local AIUtils = import('/lua/ai/aiutilities.lua')
    
### Start of Shield Projection Unit(SuperClass)
function ShieldProjectionUnit(SuperClass) 
    return Class(SuperClass) {
    
    ShieldEffects = {
        '/effects/emitters/aeon_shield_generator_mobile_01_emit.bp',
        '/effects/emitters/seraphim_regenerative_aura_01_emit.bp',
    },    
      
    OnCreate = function(self,builder,layer)
        ### Store the SPU blueprint for later use              
        self.UnitBP = self:GetBlueprint()       
        self.SPU_Econ = self.UnitBP.Defense.ProjectedShield.EcostPerUnit or 10     
               
        ### Sets up ExcludedCats
        local bpcats = self.UnitBP.Defense.ProjectedShield.CatsDisallow
        self.ExcludedCats = ParseEntityCategory(bpcats)            
                 
        ### Sets up tables to track the units being shielded         
        self.ShieldUnitTable = {}
        
        ### Table for the required shield effects
        self.ShieldEffectsBag = {}
        SuperClass.OnCreate(self,builder,layer)
    end,
    
    OnKilled = function(self, instigator, type, overkillRatio)
        if self:GetShieldStatus() then 
            self:SetShieldStatus(false)
        end
        SuperClass.OnKilled(self, instigator, type, overkillRatio)
    end,    
    
    OnScriptBitSet = function(self, bit)
        SuperClass.OnScriptBitSet(self, bit)
        if bit == 0 then 
            if not self:GetShieldStatus() then 
                self:PlayUnitSound('ShieldOn')
                self:SetShieldStatus(true)
                
                if self.ShieldEffectsBag then
                    for k, v in self.ShieldEffectsBag do
                        v:Destroy()
                    end
                    self.ShieldEffectsBag = {}
		        end
                for k, v in self.ShieldEffects do
                    table.insert( self.ShieldEffectsBag, CreateAttachedEmitter( self, 0, self:GetArmy(), v ) )
                end                               
            end                        
        end
    end,
    
    OnScriptBitClear = function(self, bit)   
        SuperClass.OnScriptBitClear(self, bit)
        if bit == 0 then
            if self:GetShieldStatus() then 
                self:PlayUnitSound('ShieldOff')
                self:SetShieldStatus(false)
                
                if self.ShieldEffectsBag then
                    for k, v in self.ShieldEffectsBag do
                        v:Destroy()
                    end
		            self.ShieldEffectsBag = {}
                end                
            end
        end
    end,                  
    
    GetShieldStatus = function(self)
        return self.AreaShieldActive
    end,  
    
    SetShieldStatus = function(self, enabled)
        ### Update the global variable
        self.AreaShieldActive = enabled
        if enabled then  
            ### Add or enable shields   
            self:ForkThread(self.AreaScan)
        else
            ### Disable shields    
            if table.getn(self.ShieldUnitTable) > 0 then
                while table.getn(self.ShieldUnitTable) > 0 do
                    self:ShieldDisable(self.ShieldUnitTable[1])
                end
            end  
        end
    end,        
    
    AreaScan = function(self)
        ### Small delay before the shield kicks on
        WaitTicks(1)
        while not self:IsDead() and self:GetShieldStatus() do

            ### Check for potential units near SPU
            local unitsInArea = AIUtils.GetOwnUnitsAroundPoint(self:GetAIBrain(), categories.ALLUNITS - self.ExcludedCats, self:GetPosition(), self.UnitBP.Defense.ProjectedShield.EffectRadius) 
            
            ### Check that units have been detected nearby and that the SPU is not in transport
            if table.getn(unitsInArea) > 0 and not self:IsUnitState('Attached') then
                           
                if table.getn(self.ShieldUnitTable) > 0 then 
                    ### Disable shielding if unit moves out of the area
                    for k, v in self.ShieldUnitTable do
                        if not table.find(unitsInArea, v) then
                            self:ShieldDisable(v)
                        end
                    end 
                end 
                
                for k, v in unitsInArea do
                    ### Check for units in the area that are not already shielded
                    if not table.find(self.ShieldUnitTable, v) then
                        if self:GetShieldStatus() and not v:ShieldIsOn() then
                            self:ShieldEnhance(v)                           
                        end
                    end 
                end                  
                            
            elseif table.getn(self.ShieldUnitTable) > 0 then
                ### Disable all shielding as no units are in range or SPU is on transport               
                for k, v in self.ShieldUnitTable do
                    self:ShieldDisable(v)
                end 
            end                                                                                                                                                           

            ### Reset local variables
            unitsInArea = nil

            ### Delay between checks
            WaitTicks(Random(3,5))                     
        end
    end,
    
    ShieldEnhance = function(self, sUnit)
        if not self:IsDead() and not sUnit:IsDead() then 
            self:AdjEnergyCost(sUnit, true)          
            ### Activate or add shield and update the global table
            if sUnit.MyShield then
                sUnit:EnableShield()
            else
                local OverKill = self.UnitBP.Defense.ProjectedShield.PassOverkillDamage or false
                local Impact = self.UnitBP.Defense.ProjectedShield.ImpactEffects
                local RegenDelay = self.UnitBP.Defense.ProjectedShield.ShieldRegenStartTime
                sUnit:SpawnProjectedShield({self}, OverKill, Impact, RegenDelay)
            end
            table.insert(self.ShieldUnitTable, sUnit)      
        end
    end, 

    ShieldDisable = function(self, sUnit)
        if not sUnit:IsDead() and sUnit.MyShield then        
            ### Remove shield and update the global table       
            self:AdjEnergyCost(sUnit, false)                             
            sUnit:DisableShield()
        end
        table.removeByValue(self.ShieldUnitTable, sUnit)                         
    end, 
    
    AdjEnergyCost = function(self, sUnit, enabled)
        if not self:IsDead() and not sUnit:IsDead() then
                   
            local eCost
            local sUnit_bp = sUnit:GetBlueprint()
            local sUnit_Econ = sUnit_bp.Economy.MaintenanceConsumptionPerSecondEnergy or 0 
            local sUnit_Toggle = sUnit_bp.General.ToggleCaps
            if sUnit_Toggle then
                #LOG('toggle caps: ', sUnit_Toggle)   
               if self:GetScriptBit(sUnit_Toggle) == 1 then
                   #LOG('True')
               else
                   #LOG('False')
               end
            end
            
            if sUnit_Econ > 0 then
                local e_Consumption = sUnit:GetConsumptionPerSecondEnergy()               
                if enabled then
                    eCost = self.SPU_Econ + e_Consumption
                else
                    eCost = self.SPU_Econ - e_Consumption
                end                                                             
            else
                if enabled then
                    eCost = self.SPU_Econ 
                else
                    eCost = 0
                end
            end
            
            if eCost < 0 then
                eCost = 0
            end                 
                                        
            sUnit:SetMaintenanceConsumptionInactive()  
            sUnit:SetEnergyMaintenanceConsumptionOverride(eCost)                       
            sUnit:SetConsumptionPerSecondEnergy(eCost)
            sUnit:SetMaintenanceConsumptionActive()           
        end
    end,  
                 
}        
end 
### End of Shield Projection Unit(SuperClass) ###